/*
 * error.c --  error handling routines.
 *
 * This file is part of
 * 
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1998 Lorenzo Vicisano
 *     (vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#if 0
#include <malloc.h>
#endif
#include "error.h"
#include "ui.h"
#include "agent.h"
#include "rlcif.h"

static void default_error(int pi, int errcode);

static error_handle_t error_handle = default_error;
static char msgbuf[512];
static int code = 0;
static int msg_append = 0;
static int this_pi = -1;

static char *error_msgs[] = {
    "* Memory allocation ERROR",		/*ALLOC_ERROR*/
    "* Timer allocation ERROR",			/*ALLOC_TIMIT_ERROR*/
    "* Malformed address ERROR",		/*BAD_ADDRESS*/
    "* Not in class D address ERROR",		/*ADDR_NOT_IN_CLASSD*/
    "* Bad Callback function ERROR",		/*BAD_CALLBACK_FUNCTION*/
    "* Too many protocol instances ERROR",	/*TOO_MANY_INSTANCES*/
    "* Cannot start sender ERROR",		/*CANT_START_SENDER*/
    "* Cannot keep transmitting ERROR",		/*STOP_TRANSMISSION*/
    "* Cannot allocate event list ERROR",	/*NO_EVLIST*/
    "* Bad TTL ERROR",				/*BAD_TTL*/
    "* Bad number of layers ERROR",		/*BAD_NUMBER_OF_LAYS*/
    "* Bad minimum payload size ERROR",		/*BAD_MIN_PAYLOAD*/
    "* Bad maximum payload size ERROR",		/*BAD_MAX_PAYLOAD*/
    "* Bad interpacket transmission period ERROR",	/*BAD_PERIOD*/
    "* This address is already in use in this sender host ERROR",/*ADDR_IN_USE*/
    "* Too many IO descriptor ERROR",		/*TOO_MANY_IOS*/
    "* No input channel available ERROR",	/*NO_CHAN*/
    "* Cannot start receiver ERROR",		/*CANT_START_RECEIVER*/
    "* join layer ERROR",			/*JOIN_ERROR*/
    "* Too many sockets (check MAX_RX_SOCKS) ERROR",	/*TOO_MANY_SOCKET*/
    "* Error in a network routine ERROR",	/*NET_CALL_ERROR*/
    "* leave layer ERROR",			/*LEAVE_ERROR*/
    "* attmp to transmit a too large packet ERROR",	/*TOO_LARGE_PACKET*/
    "* init TCL UI ERROR",			/*TCL_UI_ERROR*/
    "* bandwidth distribution option  not implemented ERROR", /*BAND_NOT_IMPL*/
    "* option edge trigger not implemented ERROR",	/*TRIG_NOT_IMPL*/
    "* Cannot allocate storage for incoming packet ERROR", /*ALLOC_PKT_ERROR*/
    "* Cannot allocate temporary storage ERROR",	/*ALLOC_TMP_ERROR*/
    "* Timeout in reception ERROR",	/*ALLOC_TMP_ERROR*/
    "* Cannot set watchdog ERROR",	/*ALLOC_TMP_ERROR*/
    "* Non existent protocol instance ERROR",	/*ALLOC_TMP_ERROR*/
    "* Can't change address when the protocol has started ERROR",
		/*CANT_CHANGE_ADDR*/
    "* Don't hace GUI stuff ERROR"
};

static char *warn_msgs[] = {
    "* Warning: no user interface",		/*NOUI_WARNING*/
    "* Warning: Cannot keep running UI",	/*STOP_UI*/
    "* Warning: reception error",		/*RCV_ERROR*/
    "* Warning: sending error",			/*SND_ERROR*/
    "* Warning: binding to M-cast source address failed, trying INADDRANY",
						/*MULTI_BIND_ERROR*/
    "* Warning: invalid odd port number, using the preceeding even value",
						/*ODD_PORT*/
    "* Warning: no default error handler",	/*NODEFERR_WARNING*/
    "* Warning: no handler for SIGINT",		/*NOSIGINT_WARNING*/
    "* Warning: reception timeout"		/*RECEPTION_TIMEOUT_WAR*/
};

static err_inf_t *err_ta = NULL;
static err_inf_t *warn_ta = NULL;

static int err_disp (int code, err_inf_t *e)
{
    if (code > 0) {
	if (code >= e->base && code < e->base + e->length)
	    return (code - e->base);
	return (-1);
    } else {
	if (code <= e->base && code > e->base - e->length)
	    return (e->base - code);
	return (-1);
    }
}

static char *err_msg (int code)
{
  err_inf_t **pt;

    if (IS_RLC_ERR(code))
	return (error_msgs[ERR_MSG_IDX(code)]);
    if (IS_RLC_WARN(code))
	return (warn_msgs[ERR_MSG_IDX(code)]);

    if (code < 0)
	pt = &err_ta;
    else
	pt = &warn_ta;

    while (*pt != NULL && err_disp(code, *pt) < 0)
	pt = &((*pt)->next);

    if (err_disp(code, *pt) < 0)
	return (NULL);

    return ((*pt)->table[err_disp(code, *pt)]);
}

int rlc_register_error_table (char **table, int base, int length)
{
  err_inf_t *t, **pt;
    if ( (t = (err_inf_t *)malloc(sizeof(err_inf_t))) == NULL)
	return (rlc_internal_set_error_ (-1, ALLOC_ERROR, NULL));

    if (base < 0)
	pt = &err_ta;
    else
	pt = &warn_ta;

    while (*pt != NULL)
	pt = &((*pt)->next);

    *pt = t;

    t->next = NULL;
    t->base = base;
    t->length = length;
    t->table = table;

    return (0);
}

/*++++
 Returns the code and description of the last generated error.

 Params:
	msg : pointer to a char * that will be set to point to
		the error description string.
 Returns:
	the error code
*/
int rlc_error_info(char **msg)
{
    if (msg)
	*msg = msgbuf;
    return (code);
}

/*++++
 Registers the error handler client function.  This function will be
 called when an internal exception occurs:  error/warning that
 verifies in a function called by the event manager (typically
 reception/transmission routines).
 
 The rlc client can retrieve the event code using rlc_error_info(),
 which returns a value > 0 for warning and < 0 for errors. Errors
 must be handled by the client, while warnings can be ingnored.
 
 If the client has not register the error handler, a SIGUSR1 signal is
 raised on error; warnings are ignored. A default signal handler is
 installed for SIGUSR1, that prints the error stack and terminates the
 program.

 A complete list of error/warming codes definition is in rlcif.h .

 Params:
	f : the client provided error-handling function.
		the type error_handle_t is defined as:
			typedef void (*error_handle_t) (int pi);
*/

void rlc_set_error_h(error_handle_t f)
{
    error_handle = f;
}

/*
** The actual storing function. Gets called by the function below or
** by the library to store intermediate error messages.
*/
int rlc_internal_set_error_ (int pi, int errcode, char const *fmt, ...)
{
  va_list ap;
  char *p;
    if (fmt) {
	va_start(ap, fmt);
	if (msg_append) {
#if 0
	    strncat (msgbuf, msg, sizeof(msgbuf)-1-strlen(msgbuf));
#else
	    p = msgbuf + strlen(msgbuf);
	    vsprintf(p, fmt, ap);
#endif
	} else
#if 0
	    strncpy (msgbuf, msg, sizeof(msgbuf)-1);
#else
	    vsprintf(msgbuf, fmt, ap);
#endif
    } else /* XXX shell we check ? if (HAVE_ERR_MSG(errcode)) */ {
	if (msg_append)
	    strncat (msgbuf, err_msg(errcode), sizeof(msgbuf)-1-strlen(msgbuf));
	else
	    strncpy (msgbuf, err_msg(errcode), sizeof(msgbuf)-1);
    }
    code = errcode;
    msg_append = 1;
    this_pi = pi;
    return code;
}

/*
** Stores the error message and error code in temporary variables,
** so that can be retrieved by the client using rlc_error_info().
*/
int rlc_return_error_ (int pi, int errcode, char const *fmt, ...)
{
  va_list ap;
    va_start(ap, fmt);
    rlc_internal_set_error_(pi, errcode, fmt, ap);
    msg_append = 0;
    return errcode;
}

/*
** Propagetes an error to the client calling the user-provided error
** handling function (of type *error_handle_t). If the error handling
** function is not defined, it rises a SIGUSR1 signal. The user
** can install its error handler using rlc_set_error_h(), and get the
** error cause using rlc_error_info().
*/
void rlc_have_error_ (int pi, int errcode, char const *fmt, ...)
{
  va_list ap;
    va_start(ap, fmt);
    NDEB(fprintf(stderr, "*** rlc_have_error_ for %d with (%d)\n", pi,
	errcode);)
    NDEB(if (msg) fprintf(stderr, "\t%s\n", msg);)
    (void) rlc_return_error_ (pi, errcode, fmt, ap);
    if (error_handle)
	error_handle (pi, errcode);
    else if (RLC_IS_ERR(errcode))
#if (!defined(Linux))
#if ( defined(SunOS) && (OSMVER < 5))
	kill (getpid(), SIGUSR1);    
#else
	raise (SIGUSR1);    
#endif
#else
	raise (SIGUSR1);    
#endif
}

/*
** Just print the error message stack and exit
*/
static void
default_error(int pi, int errcode)
{
  char *c;
    (void)rlc_error_info(&c);
    DEB(fprintf(stderr, "*** default_error with %s\n", c);)
    if (pi >= 0 || this_pi >= 0) {
#ifdef HAVE_UI
	sxs_t *sxs;
#endif
	int tpi;
	if (pi >= 0) tpi = pi;
	else tpi = this_pi;
#ifdef HAVE_UI
	sxs = (sxs_t *)_rlc_getinstance(tpi, rlc_inst_mngr);
	DEB(fprintf(stderr, "*** default_error pi =  %d\n", tpi);)
	if (sxs->error_gui && RLC_IS_ERR(errcode)) {
	    DEB(fprintf(stderr, "*** default_error going "
		"error gui\n");)
	    sxs->error_gui (tpi, c);
	    return;
	} else if (sxs->warning_gui && !RLC_IS_ERR(errcode)) {
	    DEB(fprintf(stderr, "*** default_error going "
		"warning gui\n");)
	    sxs->warning_gui (tpi, c);
	    return;
	}
#endif
	if (RLC_IS_ERR(errcode)) {
	    fprintf (stderr, "+++rlc (pi %d) fatal error:\n\t%s\n", tpi, c);
	    exit (1);
	} else {
	    fprintf (stderr, "+++rlc (pi %d) warning:\n\t%s\n", tpi, c);
	}
    } else {
	if (RLC_IS_ERR(errcode)) {
	    fprintf (stderr, "+++rlc fatal error:\n\t%s\n", c);
	    exit(1);
	} else {
	    fprintf (stderr, "+++rlc warning:\n\t%s\n", c);
	}
    }
}

int
_rlc_error_gui_init (int pi)
{
#ifdef HAVE_UI
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));
    if (sxs->have_gui)
	return (rlc_return_error_ (pi, NO_GUI_ERROR, NULL));
    sxs->error_gui = rlc_gui_error_handle;
    sxs->warning_gui = rlc_gui_warning_handle;
    return (0);
#else
    return (rlc_return_error_ (pi, NO_GUI_ERROR, NULL));
#endif
}

/*
** Install the default error-signal handler.
*/
int set_default_error_h ()
{
    rlc_set_error_h (default_error);
#if 0
    if (signal(SIGUSR1, default_error) < 0) {
	return (rlc_internal_set_error_ (-1, NODEFERR_WARNING,
	    strerror (errno)));
    }
#endif
    return (0);
}
