/* 
 * layers.c -- layers handling routines.
 *
 * This file is part of
 *
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include <string.h>
#include "rlc.h"
#include "rlcif.h"
#include "layers.h"
#include "agent.h"
#include "net.h"
#include "utils.h"

/*
 * Setup trasmission sockets for layers,
 * assume addr is for layer 0 and layer
 * `i' has address addr+i.
 */
int
setup_l_sock(lay_d *l, addr_list_t *addr, n16 port, int ttl)
{
    int i, consecutive;
    ui32 tmpaddr[MAX_LAYERS];

    if ((consecutive = set_addresses(addr, l->nc, tmpaddr)) < 0)
	return (rlc_internal_set_error_(-1, consecutive, NULL));

#ifdef ONE_SEND_SOCK
    /* if called with unbound=1 port can be 0 */
    if ((l->chan = openssock(tmpaddr[0], port, ttl, 1)) < 0)
	return (rlc_internal_set_error_ (-1, l->chan, NULL));
    NDEB(fprintf(stderr, "--LAY: have a new sock: %s %d + 2*i, ttl %d\n",
	inet_ntoa(*(struct in_addr *)&(tmpaddr[0])), port, ttl))
#endif
    for (i=0; i<(l->nc); i++) {
#ifdef ONE_SEND_SOCK
	bzero((char *)&(l->saddrs[i]), sizeof(l->saddrs[i]));
	l->saddrs[i].sin_family = AF_INET;
	l->saddrs[i].sin_addr.s_addr = tmpaddr[i];
	l->saddrs[i].sin_port = htons(ntohs(port) + 2 * i);
#else
	if ((l->chs[i] = openssock(tmpaddr[i], htons(ntohs(port) + 2 * i),
	    ttl, 0)) < 0)
	    return (rlc_internal_set_error_ (-1, l->chs[i], NULL));
	NDEB(fprintf(stderr, "--LAY: have a new sock: %s %d, ttl %d\n",
		inet_ntoa(*(struct in_addr *)&(tmpaddr[i])),
		htons(ntohs(port) + 2 * i), ttl))
#endif
    }
    return consecutive;
}

ui32
l_change_rate(ui32 rate, lay_d *l)
{
    l->pu = rate;

    NDEB(
    for (i=(l->nc)-1; i>=0; i--) {
	fprintf(stderr, "*** lay %d period = %d (%d), next = %d (%d)\n",
		i, l->rates[i], l->rates[i]*l->pu, l->left[i],
		l->left[i]*l->pu);
    }
    )

    return(rate);
}

/*
 * Setup data stuctures for layers.
 * `rate' is the inverse of the overall
 * packet rate.
 */
void
setup_layers(ui32 rate, lay_d *l, int W, int Y, int P)
{
    int i;
    ui32 base;
    int t_W = W*(1<<(l->nc-1));
    int t_Y = Y*(1<<(l->nc-1));

    /* compute a random pn ... multiple of W */
#if 0		/** XXX to debug.. there might be problems! **/
    l->pn = (ui16) (timestamp_u()*W);
#else
    l->pn = 0;
#endif

    l->pu = rate;
    l->inc_todo = 0;
    l->W = W;
    l->Y = Y;
    l->P = P;
    l->now = 0;				/* debugging only */

    base = rate = 1;
    for (i=(l->nc)-1; i>=0; i--) {
	l->left[i] = rate - base;
	if (i != 0) {
	    rate *= 2;
	    t_W /= 2;
	    t_Y /= 2;
	}
	l->rates[i] = rate;
	NDEB(fprintf(stderr, "*** lay %d period = %d (%d), next = %d (%d)\n",
		i, rate, rate*l->pu, l->left[i], l->left[i]*l->pu))
	l->idxs[i] = l->pn * (1<<max(0, i-1));
	l->cyc_l[i] = t_W;
	l->bur_l[i] = t_Y;
	l->gap[i] = l->rates[i] * ((l->bur_l[i] >> 1) + 1);
    }
}


/*
 * Establish where to send now (lay) and
 * return the time to next send.
 * We assume that at least one `left' is = 0
 *
 * BURSTS should be arranged here !!
 *
 */
static ui32
next_to_send(lay_d *l, int *lay, ui16 *lidx, char *f)
{
    int i;
    int app;
    ui32 minor = 0xffffffff;

    *lay = -1;

    /* search which layer to trasmit on
     * (precedence is given to faster layers)
     */
    for (i=(l->nc)-1; i>=0; i--) {
	if (*lay == -1 && l->left[i] == 0) {
	    *lay = i;
	    /*
	     * Layer `i' has been selected, now set
	     * left to next trasm. time in this layer.
	     * To introduce burst change below, lowering
	     * left, when appropriate.
	     */
	    app = l->idxs[i] % l->cyc_l[i];
	    if (app < l->bur_l[i]) {		/* this is in burst */
		*f |= RLC_F_IN_BURST;
		if (!app)
		    *f |= RLC_F_S_F_BURST;
		if (app == l->bur_l[i]-1) {	/* this is the last */
		    if (!i)
			*f |= RLC_F_L_BURST;
		    l->left[i] = l->gap[i];
		} else {
		    if (!(app%2)) {		/* next is attached */
			l->left[i] = 0;
		    } else {
			l->left[i] = l->rates[i];
		    }
		}
	    } else {				/* non in burst */
		if (app == l->bur_l[i])
		    *f |= RLC_F_S_F_BURST;
		l->left[i] = l->rates[i];
	    }
	}
	minor = min(minor, l->left[i]);
    }

    /* adjust times */
    if (minor)
	for (i=0; i<(l->nc); i++)
	    l->left[i] = l->left[i] - minor;

    *lidx = l->idxs[*lay];
    (l->idxs[*lay]) ++;
    
    return (minor*l->pu);
}

char *trst[] = {"normal pkt", "no data in buff.", "skip to sync",
        "burst", "frame over"};

/*
 * Establish where to send next packet
 * (lay, lidx) and get its ponter p and
 * lenght. Return the time to next send.
 */
ui32
send_layers(sxs_t *s, char **p, ui32 *len, int *lay, ui16 *pn, ui16 *lidx,
	char *f)
{
    int res;
    ui32 delta;
    char flags = '\0';
    lay_d *l = &(s->l);

    delta = next_to_send(&(s->l), lay, lidx, f);

 /* adjust the indexes like that:
 **
 **     0,1   2,3   4,5   6,7                      8     9     10    11    12
 **	0,0   0,0   0,0   0,0                      2     2     2     2     3
 **     XX    XX    XX    XX                       X     X     X     X     X
 **        0,1         2,3                            4           5           6 
 **        0,0         0,0                            2           2           3
 **        XX          XX                             X           X           X
 **              0,1                                        3 
 **              0,0                                        2
 **              XX                                         X
 **                           0,1                                       2 
 **                           0,0                                       2
 **                           XX                                        X
 */

    *pn = l->pn;
    if ( !(*lay) )
	if ( !(l->pn%l->W) ) {	/* not increment at the first on
							   the burst */
	    if (!(l->inc_todo))
		l->inc_todo = 1;
	    else {
		l->inc_todo = 0;
		l->pn += 2;
	    }
	} else
	    l->pn ++;

    /* mark up sync. point */
    if (*f & RLC_F_IN_BURST && ( (*pn/l->W )%((1<<*lay)*l->P) == 0 ))
	*f |= RLC_F_P_BURST;

    res = s->tx_packet(s->cpi, *lay, /* *lidx,*/ flags, p);
    if (res > 0) {
	*len = res;
	res = 0;
    } else if (res < 0) {
	*len = res;
	res = -res;
    } else {
	*len = 0;
	return(IS_OVER);
    }
    NDEB(fprintf(stderr, "send_layers: %d bytes for %s (%d, %d)\n",
	    *len, trst[res], *lay, *lidx));

    NDEB(fprintf(stderr, "+++ L: we had lay %d lidx %d\n",
	*lay, *lidx))

    return (delta);
}

/*
 * Compute a global index on 32 bits (wrapping) on the basis of
 * the current joined layer (curr_lay). Used in the receiver to detect
 * gaps and to reorder packets.
 */
ui32	global_index (ui16 lsn, ui16 pn, int lay, int curr_lay, int inbu)
{
    if (inbu) {	/* IN_BURST */
#if 0
ui32 a = ((lsn/2) % (1<<max(lay-1,0)));
ui32 b = (1<< (curr_lay-max(lay,1)+1));
ui32 c = (1<<(curr_lay-lay)) - 1;
	fprintf(stderr, "**** have a %lu b %lu c %lu\n", a, b, c);
#endif
	return(
	(ui32)pn * (1<<curr_lay) +
	2 * ( ((lsn/2) % (1<<max(lay-1,0))) * (1<< (curr_lay-max(lay,1)+1))
	+ (1<<(curr_lay-lay)) - 1 ) +
	lsn % 2 );
    } else
	return(
	(ui32)pn * (1<<curr_lay) +
	(lsn % (1<<max(lay-1,0))) * (1<< (curr_lay-max(lay,1)+1))
	+ (1<<(curr_lay-lay)) - 1);
}

/*
 * Do the inverse of the function above: Compute lsn, pn and lay given
 * gidx on the basis of the current joined layer (curr_lay).
 */
void	from_global_index (ui32 gidx, ui16 *lsn, ui16 *pn, int *lay, 
	int curr_lay, int inbu)
{
  int i, off;

    *pn = gidx / (1<<curr_lay);
    
    if (inbu) {		/* in burst .. */
	off = gidx % (1<<(curr_lay+1));
	off >>= 1;
    } else {
	off = gidx % (1<<curr_lay);
    }

    for (i = curr_lay; i > 0; i--) {
	if (!(off % 2))
	    break;
	off >>= 1;
    }
    *lay = i;

    *lsn = *pn * (1<<(max(*lay-1, 0)+inbu)) + inbu * gidx%2;

    if (inbu)
	*pn -= *pn % 2;
}

int is_in_burst (char f)
{
    if (f&RLC_F_IN_BURST)
	return (1);
    return (0);
}

int is_first_burst (int lay, int curr_lay, char f)
{
    if (lay == curr_lay && is_in_burst(f) && (f&RLC_F_S_F_BURST))
	return (1);
    return (0);
}

int is_first_non_burst (int lay, int curr_lay, char f)
{
    if (lay == curr_lay && !is_in_burst(f) && (f&RLC_F_S_F_BURST))
	return (1);
    return (0);
}

int is_burst_ending (char f)
{
    if (f&RLC_F_L_BURST)
	return (1);
    return (0);
}

int up_synch_burst (int lay, int curr_lay, char f)
{
    if (lay == curr_lay && f&RLC_F_P_BURST)
	return (1);
    return (0);
}
